# -*- coding: utf-8 -*-
import sys
import random
from shutil import rmtree
from os.path import join
import unittest
from skoolkittest import SkoolKitTestCase
from skoolkit.skoolhtml import HtmlWriter, FileInfo, Udg
from skoolkit.skoolparser import SkoolParser, RefParser

TEST_SKOOL = 'test-html.skool'
TEST_REF = 'test-html.ref'
ODIR = 'html'
GAMEDIR = 'test'
ASMDIR = 'asm'
REFERENCEDIR = 'reference'
UDGDIR = 'images/udgs'

TABLE_SRC = """(data)
{ =h Col1 | =h Col2 | =h,c2 Cols3+4 }
{ =r2 X   | Y       | Za  | Zb }
{           Y2      | Za2 | =t }
"""

TABLE_HTML = """
<table class="data">
<tr>
<th>Col1</th>
<th>Col2</th>
<th colspan="2">Cols3+4</th>
</tr>
<tr>
<td rowspan="2">X</td>
<td>Y</td>
<td>Za</td>
<td>Zb</td>
</tr>
<tr>
<td>Y2</td>
<td>Za2</td>
<td class="transparent"></td>
</tr>
</table>
"""

class TestHtmlWriter(HtmlWriter):
    def write_image(self, img_file, udgs=(), scale=2, mask=False, x=0, y=0, width=None, height=None):
        self.image_writer.write_image(udgs, scale, mask, x, y, width, height)

class MockSkoolParser:
    def __init__(self, snapshot, entries=(), memory_map=()):
        self.snapshot = snapshot
        self.entries = entries
        self.memory_map = memory_map
        self.skoolfile = ''

class MockImageWriter:
    def __init__(self):
        self.build_images = True

    def write_image(self, udg_array, scale, mask, x, y, width, height):
        self.udg_array = udg_array
        self.scale = scale
        self.mask = mask
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def get_default_format(self):
        return 'png'

class HtmlWriterTest(SkoolKitTestCase):
    def read_file(self, fname):
        f = open(join(ODIR, GAMEDIR, fname), 'r')
        contents = f.read()
        f.close()
        return contents

    def _test_reference_macro(self, macro, def_link_text, page):
        writer = self._get_writer()
        for link_text in ('', '(test)', '(test (nested) parentheses)'):
            for anchor in ('', '#test'):
                output = writer.expand('#%s%s%s' % (macro, anchor, link_text), ASMDIR)
                self.link_equals(output, '../%s/%s.html%s' % (REFERENCEDIR, page, anchor), link_text[1:-1] or def_link_text)

    def _test_call(self, cwd, arg1, arg2):
        # Method used to test the #CALL macro
        return '%s%s' % (cwd, arg2 * arg1)

    def _get_writer(self, use_ref=False, snapshot=None):
        ref_parser = RefParser()
        if use_ref:
            ref_parser.parse(TEST_REF)
        if snapshot:
            skool_parser = MockSkoolParser(snapshot)
        else:
            skool_parser = SkoolParser(TEST_SKOOL, html=True)
        file_info = FileInfo(ODIR, GAMEDIR, False)
        return TestHtmlWriter(skool_parser, ref_parser, file_info, MockImageWriter())

    def _assert_scr_equal(self, game, x0=0, y0=0, w=32, h=24):
        snapshot = game.snapshot[:]
        scr = game.get_screenshot(x0, y0, w, h)
        self.assertEqual(len(scr),  min((h, 24 - y0)))
        self.assertEqual(len(scr[0]), min((w, 32 - x0)))
        for j, row in enumerate(scr):
            for i, udg in enumerate(row):
                x, y = x0 + i, y0 + j
                df_addr = 16384 + 2048 * (y // 8) + 32 * (y % 8) + x
                af_addr = 22528 + 32 * y + x
                self.assertEqual(udg.data, snapshot[df_addr:df_addr + 2048:256], 'Graphic data for cell at (%i,%i) is incorrect' % (x, y))
                self.assertEqual(udg.attr, snapshot[af_addr], 'Attribute byte for cell at (%i,%i) is incorrect' % (x, y))

    def link_equals(self, html, href, text):
        self.assertEqual(html, '<a class="link" href="%s">%s</a>' % (href, text))

    def img_equals(self, html, alt, src):
        self.assertEqual(html, '<img alt="%s" src="%s" />' % (alt, src))

    def _check_image(self, image_writer, udg_array, scale, mask, x, y, width, height):
        self.assertEqual(image_writer.scale, scale)
        self.assertEqual(image_writer.mask, mask)
        self.assertEqual(image_writer.x, x)
        self.assertEqual(image_writer.y, y)
        self.assertEqual(image_writer.width, width)
        self.assertEqual(image_writer.height, height)
        self.assertEqual(len(image_writer.udg_array), len(udg_array))
        for i, row in enumerate(image_writer.udg_array):
            exp_row = udg_array[i]
            self.assertEqual(len(row), len(exp_row))
            for j, udg in enumerate(row):
                exp_udg = exp_row[j]
                self.assertEqual(udg.attr, exp_udg.attr)
                self.assertEqual(udg.data, exp_udg.data)
                self.assertEqual(udg.mask, exp_udg.mask)

    def test_get_screenshot(self):
        snapshot = [0] * 65536
        for a in range(16384, 23296):
            snapshot[a] = random.randint(0, 255)
        writer = self._get_writer(snapshot=snapshot)
        self._assert_scr_equal(writer)
        self._assert_scr_equal(writer, 1, 2, 12, 10)
        self._assert_scr_equal(writer, 10, 10)

    def test_get_font_udg_array(self):
        snapshot = [0] * 65536
        char1 = [1, 2, 3, 4, 5, 6, 7, 8]
        char2 = [8, 7, 6, 5, 4, 3, 2, 1]
        chars = [char1, char2]
        char_data = []
        for char in chars:
            char_data.extend(char)
        address = 32768
        snapshot[address:address + sum(len(c) for c in chars)] = char_data
        writer = self._get_writer(snapshot=snapshot)
        attr = 56
        width, font_udg_array = writer.get_font_udg_array(address, len(chars), attr)
        self.assertEqual(width, 8 * len(chars))
        self.assertEqual(len(font_udg_array[0]), len(chars))
        for i, udg in enumerate(font_udg_array[0]):
            self.assertEqual(udg.attr, attr)
            self.assertEqual(udg.data, chars[i])

    def test_get_instruction_addr_str(self):
        writer = self._get_writer()
        self.assertEqual(writer.get_instruction_addr_str(24576), '24576')
        self.assertEqual(writer.get_instruction_addr_str(24579), '6003')

    def test_ref_parsing(self):
        writer = self._get_writer(use_ref=True)
        self.assertEqual(writer.info, {'X': 'foo', 'Y': 'bar'})

    def test_macro_bug(self):
        self._test_reference_macro('BUG', 'bug', 'bugs')

    def test_macro_c(self):
        writer = self._get_writer()
        output = writer.expand('#C', ASMDIR)
        self.assertEqual(output, '&#169;')

    def test_macro_call(self):
        writer = self._get_writer()
        writer.test_call = self._test_call
        output = writer.expand('#CALL:test_call(10,test)', ASMDIR)
        self.assertEqual(output, self._test_call(ASMDIR, 10, 'test'))

    def test_macro_d(self):
        writer = self._get_writer()
        output = writer.expand('#D24592', ASMDIR)
        self.assertEqual(output, 'The final routine')

    def test_macro_erefs(self):
        writer = self._get_writer()
        output = writer.expand('#EREFS24582', ASMDIR)
        self.assertEqual(output, 'routines at <a class="link" href="24586.html">24586</a> and <a class="link" href="24592.html">24592</a>')

    def test_macro_fact(self):
        self._test_reference_macro('FACT', 'fact', 'facts')

    def test_macro_font(self):
        snapshot = [0] * 65536
        writer = self._get_writer(snapshot=snapshot)

        output = writer.expand('#FONT32768,96', ASMDIR)
        self.img_equals(output, 'font', '../images/font/font.png')

        img_fname = 'font2'
        output = writer.expand('#FONT55584,96(%s)' % img_fname, ASMDIR)
        self.img_equals(output, img_fname, '../images/font/%s.png' % img_fname)

        img_fname = 'font3'
        font_addr = 32768
        char1 = [1, 2, 3, 4, 5, 6, 7, 8]
        char2 = [8, 7, 6, 5, 4, 3, 2, 1]
        char3 = [1, 3, 7, 15, 31, 63, 127, 255]
        chars = (char1, char2, char3)
        for i, char in enumerate(chars):
            snapshot[font_addr + i * 8:font_addr + (i + 1) * 8] = char
        attr = 4
        scale = 2
        x, y, w, h = 1, 2, 3, 4
        macro = '#FONT%i,%i,%i,%i{%i,%i,%i,%i}(%s)' % (font_addr, len(chars), attr, scale, x, y, w, h, img_fname)
        output = writer.expand(macro, ASMDIR)
        self.img_equals(output, img_fname, '../images/font/%s.png' % img_fname)
        udg_array = [[Udg(4, char) for char in chars]]
        self._check_image(writer.image_writer, udg_array, scale, False, x, y, w, h)

    def test_macro_link(self):
        writer = self._get_writer()

        link_text = 'bugs'
        output = writer.expand('#LINK:Bugs(%s)' % link_text, ASMDIR)
        self.link_equals(output, '../%s/bugs.html' % REFERENCEDIR, link_text)

        link_text = 'pokes'
        anchor = '#poke1'
        output = writer.expand('#LINK:Pokes%s(%s)' % (anchor, link_text), ASMDIR)
        self.link_equals(output, '../%s/pokes.html%s' % (REFERENCEDIR, anchor), link_text)

    def test_macro_name(self):
        writer = self._get_writer()
        anchor = 'testing'
        output = writer.expand('#NAME%s' % anchor, ASMDIR)
        self.assertEqual(output, '<a name="%s"></a>' % anchor)

    def test_macro_poke(self):
        self._test_reference_macro('POKE', 'poke', 'pokes')

    def test_macro_pokes(self):
        writer = self._get_writer()
        snapshot = writer.snapshot

        output = writer.expand('#POKES32768,255', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768], 255)

        output = writer.expand('#POKES32768,254,10', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768:32778], [254] * 10)

        output = writer.expand('#POKES32768,253,20,2', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768:32808:2], [253] * 20)

    def test_macro_pops(self):
        writer = self._get_writer()
        addr, byte = 49152, 128
        writer.snapshot[addr] = byte
        writer.push_snapshot('test')
        writer.snapshot[addr] = (byte + 127) % 256
        output = writer.expand('#POPS', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(writer.snapshot[addr], byte)

    def test_macro_pushs(self):
        writer = self._get_writer()
        addr, byte = 32768, 64
        writer.snapshot[addr] = byte
        output = writer.expand('#PUSHStest', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(writer.snapshot[addr], byte)
        writer.snapshot[addr] = (byte + 127) % 256
        writer.pop_snapshot()
        self.assertEqual(writer.snapshot[addr], byte)

    def test_macro_r(self):
        writer = self._get_writer(True)

        output = writer.expand('#R24576', ASMDIR)
        self.link_equals(output, '24576.html', '24576')

        link_text = 'Testing1'
        output = writer.expand('#R24576(%s)' % link_text, ASMDIR)
        self.link_equals(output, '24576.html', link_text)

        output = writer.expand('#R24576', 'other')
        self.link_equals(output, '../%s/24576.html' % ASMDIR, '24576')

        output = writer.expand('#R24579', ASMDIR)
        self.link_equals(output, '24579.html', '6003')

        output = writer.expand('#R24580', ASMDIR)
        self.link_equals(output, '24579.html#24580', '6004')

        output = writer.expand('#R32768@other', ASMDIR)
        self.link_equals(output, '../other/32768.html', '32768')

        link_text = 'Testing2'
        anchor = 'testing3'
        output = writer.expand('#R32768@other#%s(%s)' % (anchor, link_text), ASMDIR)
        self.link_equals(output, '../other/32768.html#%s' % anchor, link_text)

    def test_macro_refs(self):
        writer = self._get_writer()
        output = writer.expand('#REFS24581', ASMDIR)
        self.assertEqual(output, 'routines at <a class="link" href="24583.html">24583</a>, <a class="link" href="24586.html">24586</a> and <a class="link" href="24592.html">24592</a>')

    def test_macro_reg(self):
        writer = self._get_writer()
        for reg in ("a", "b", "c", "d", "e", "h", "l", "i", "r", "ixl", "ixh", "iyl", "iyh", "b'", "c'", "d'", "e'", "h'", "l'", "bc", "de", "hl", "sp", "ix", "iy", "bc'", "de'", "hl'"):
            output = writer.expand('#REG%s' % reg, ASMDIR)
            self.assertEqual(output, '<span class="register">%s</span>' % reg.upper())

    def test_macro_scr(self):
        snapshot = [0] * 65536
        writer = self._get_writer(snapshot=snapshot)

        output = writer.expand('#SCR', ASMDIR)
        self.img_equals(output, 'scr', '../images/scr/scr.png')

        scr_fname = 'scr2'
        output = writer.expand('#SCR2,0,0,10,10(%s)' % scr_fname, ASMDIR)
        self.img_equals(output, scr_fname, '../images/scr/%s.png' % scr_fname)

        scr_fname = 'scr3'
        data = [128, 64, 32, 16, 8, 4, 2, 1]
        attr = 48
        snapshot[16384:18432:256] = data
        snapshot[22528] = attr
        scale = 2
        x, y, w, h = 1, 2, 5, 6
        macro = '#SCR%i,0,0,1,1{%i,%i,%i,%i}(%s)' % (scale, x, y, w, h, scr_fname)
        output = writer.expand(macro, ASMDIR)
        self.img_equals(output, scr_fname, '../images/scr/%s.png' % scr_fname)
        udg_array = [[Udg(attr, data)]]
        self._check_image(writer.image_writer, udg_array, scale, False, x, y, w, h)

    def test_macro_space(self):
        writer = self._get_writer()

        output = writer.expand('#SPACE', ASMDIR)
        self.assertEqual(output, '&#160;')

        num = 10
        output = writer.expand('#SPACE%i' % num, ASMDIR)
        self.assertEqual(output, '&#160;' * num)

    def test_macro_table(self):
        writer = self._get_writer()
        output = writer.expand('#TABLE%s\nTABLE#' % TABLE_SRC, ASMDIR)
        self.assertEqual(output, TABLE_HTML)

    def test_macro_udg(self):
        snapshot = [0] * 65536
        writer = self._get_writer(snapshot=snapshot)

        udg_fname = 'udg32768_2x4'
        output = writer.expand('#UDG32768,2', ASMDIR)
        self.img_equals(output, udg_fname, '../%s/%s.png' % (UDGDIR, udg_fname))

        udg_fname = 'test_udg'
        output = writer.expand('#UDG32768,2,6,1,0:49152,2(%s)' % udg_fname, ASMDIR)
        self.img_equals(output, udg_fname, '../%s/%s.png' % (UDGDIR, udg_fname))

        udg_fname = 'test_udg2'
        udg_addr = 32768
        attr = 2
        scale = 1
        step = 1
        inc = 0
        mask_addr = 32776
        x, y, w, h = 2, 1, 3, 4
        udg_data = [136] * 8
        udg_mask = [255] * 8
        snapshot[udg_addr:udg_addr + 8 * step:step] = udg_data
        snapshot[mask_addr:mask_addr + 8 * step:step] = udg_mask
        macro = '#UDG%i,%i,%i,%i,%i:%i,%i{%i,%i,%i,%i}(%s)' % (udg_addr, attr, scale, step, inc, mask_addr, step, x, y, w, h, udg_fname)
        output = writer.expand(macro, ASMDIR)
        self.img_equals(output, udg_fname, '../%s/%s.png' % (UDGDIR, udg_fname))
        udg_array = [[Udg(attr, udg_data, udg_mask)]]
        self._check_image(writer.image_writer, udg_array, scale, True, x, y, w, h)

    def test_macro_udgarray(self):
        snapshot = [0] * 65536
        writer = self._get_writer(snapshot=snapshot)

        udg_fname = 'test_udg_array'
        output = writer.expand('#UDGARRAY8;32768-32784(%s)' % udg_fname, ASMDIR)
        self.img_equals(output, udg_fname, '../%s/%s.png' % (UDGDIR, udg_fname))

        udg_fname = 'test_udg_array2'
        output = writer.expand('#UDGARRAY8,56,2,256,0;32768;32769;32770;32771;32772x12(%s)' % udg_fname, ASMDIR)
        self.img_equals(output, udg_fname, '../%s/%s.png' % (UDGDIR, udg_fname))

        udg_fname = 'test_udg_array3'
        udg_addr = 32768
        mask_addr = 32769
        width = 2
        attr = 5
        scale = 1
        step = 256
        inc = 0
        x, y, w, h = 4, 6, 8, 5
        udg_data = [195] * 8
        udg_mask = [255] * 8
        snapshot[udg_addr:udg_addr + 8 * step:step] = udg_data
        snapshot[mask_addr:mask_addr + 8 * step:step] = udg_mask
        macro = '#UDGARRAY%i,%i,%i,%i,%i;%ix4:%ix4{%i,%i,%i,%i}(%s)' % (width, attr, scale, step, inc, udg_addr, mask_addr, x, y, w, h, udg_fname)
        output = writer.expand(macro, ASMDIR)
        self.img_equals(output, udg_fname, '../%s/%s.png' % (UDGDIR, udg_fname))
        udg_array = [[Udg(attr, udg_data, udg_mask)] * width] * 2
        self._check_image(writer.image_writer, udg_array, scale, True, x, y, w, h)

    def test_macro_udgtable(self):
        writer = self._get_writer()
        output = writer.expand('#UDGTABLE%s\nUDGTABLE#' % TABLE_SRC, ASMDIR)
        self.assertEqual(output, TABLE_HTML)

    def test_html_escape(self):
        # Check that HTML characters from the skool file are escaped
        rmtree(ODIR, True)
        writer = self._get_writer()
        fname = 'test.html'
        writer.write_entry(ASMDIR, writer.entries[24576], fname)
        html = self.read_file(join(ASMDIR, fname))
        self.assertTrue('a &lt;= b and b &gt;= c' in html)
        rmtree(ODIR, True)

    def test_html_no_escape(self):
        # Check that HTML characters from the ref file are not escaped
        rmtree(ODIR, True)
        writer = self._get_writer(True)
        writer.write_bugs()
        html = self.read_file(join(REFERENCEDIR, 'bugs.html'))
        self.assertTrue('<p>Hello</p>' in html)
        rmtree(ODIR, True)

if __name__ == '__main__':
    unittest.main()
